program jacobi_cafomp

! Solve [(d/dx)2 + (d/dy)2] u(x,y) = f(x,y) for u(x,y) in a rectangular
! domain: 0 < x < 1 and 0 < y < 1.

  implicit none
  real, allocatable :: u(:,:)[:], unew(:,:)[:], f(:,:)[:]
  integer :: ngrid         ! number of grid cells along each axis
  integer :: n             ! number of cells: n = ngrid - 1
  integer :: maxiter       ! max number of Jacobi iterations
  real    :: tol           ! convergence tolerance threshold
  real    :: omega         ! relaxation parameter
  integer i, j, k
  real    h, utmp, diffnorm
  integer np, myid
  integer js, je, js1, je1

  np   = num_images()
  myid = this_image()

! Read in problem and solver parameters.

  call read_params(ngrid,maxiter,tol,omega)

  n = ngrid - 1

! j-loop start and ending indices

  call get_indices(js,je,js1,je1,n)

! Allocate memory for arrays.

  allocate(u(0:n,js-1:je+1)[*], unew(0:n,js-1:je+1)[*], f(0:n,js:je)[*])

! Initialize f, u(0,*), u(n:*), u(*,0), and u(*,n).

  call init_fields(u,f,n,js,je)

! Main solver loop.

  h = 1.0 / n

  do k=1,maxiter
    sync all

    if (myid < np) u(:,js-1)[myid+1] = u(:,je)
    if (myid >  1) u(:,je+1)[myid-1] = u(:,js)

    sync all

!omp$ parallel do private(utmp)
    do j=js1,je1
      do i=1,n-1
        utmp = 0.25 * ( u(i+1,j) + u(i-1,j) &
             + u(i,j+1) + u(i,j-1) &
             - h * h * f(i,j) )
        unew(i,j) = omega * utmp + (1. - omega) * u(i,j)
      enddo
    enddo
!omp$ end parallel do

    call set_bc(unew,n,js,je)

!   Compute the difference between unew and u.

    call compute_diff(u,unew,n,js,je,diffnorm)

    if (myid == 1) print *, k, diffnorm

!   Make the new value the old value for the next iteration.

!omp$ parallel do
    do j=js-1,je+1
      u(:,j) = unew(:,j)
    end do
!omp$ end parallel do

!   Check for convergence of unew to u. If converged, exit the loop.

    if (diffnorm <= tol) exit
  enddo

end program

!----------------------------------------------------------------------

subroutine read_params(ngrid,maxiter,tol,omega)

  implicit none
  integer ngrid, maxiter
  real    tol, omega
  namelist /params/ ngrid, maxiter, tol, omega
  integer np, myid, i
  integer iwork(2)[*]
  real    rwork(2)[*]

  np   = num_images()
  myid = this_image()

  if (myid == 1) then
!   Default values

    ngrid   = 1000
    maxiter = 1000
    tol     = 1.e-3
    omega   = 0.75

    open(10,file='indata')
    read(10,nml=params)
    close(10)

    if (mod(ngrid,np) /= 0) then  ! For a simple CAF example code
       write(0,*) 'ERROR: ngrid must be divisible by the number of images'
       all stop
    endif
    if ((ngrid)/np < 1) then
       write(0,*) 'ERROR: local grid size should be greater than 0'
       all stop
    endif

    iwork(:) = (/ ngrid, maxiter /)
    rwork(:) = (/ tol, omega /)

    do i=2,np
      iwork(:)[i] = iwork(:)
      rwork(:)[i] = rwork(:)
    enddo
  endif

  sync all

  ngrid   = iwork(1)
  maxiter = iwork(2)
  tol     = rwork(1)
  omega   = rwork(2)

end subroutine read_params

!----------------------------------------------------------------------

subroutine get_indices(js,je,js1,je1,n)

  implicit none
  integer js, je, js1, je1, n
  integer np, myid

  np   = num_images()
  myid = this_image()

  js = 0
  je = ((n + 1) / np) - 1

  js1 = js
  if (myid ==  1) js1 = 1
  je1 = je
  if (myid == np) je1 = je - 1

end subroutine get_indices

!----------------------------------------------------------------------

subroutine init_fields(u,f,n,js,je)

  implicit none
  integer n, js, je
  real    u(0:n,js-1:je+1), f(0:n,js:je)
  integer j

! RHS term:

!omp$ parallel
!omp$ do
  do j=js,je
    f(:,j) = 4.
  end do
!omp$ end do nowait

! Initial guess:

!omp$ do
  do j=js-1,je+1
    u(:,j) = 0.5
  end do
!omp$ end do nowait
!omp$ end parallel

! Apply the boundary conditions.

  call set_bc(u,n,js,je)

end subroutine init_fields

!----------------------------------------------------------------------

subroutine set_bc(u,n,js,je)

  implicit none
  integer n, js, je
  real    u(0:n,js-1:je+1)
  integer i, j, joff, np, myid
  real    h

  np   = num_images()
  myid = this_image()
  joff = (myid - 1) * ((n + 1) / np)     ! j-index offset

  h = 1.0 / n

  if (myid == 1) then
!omp$ parallel do
    do i=0,n
      u(i,js) = (i * h)**2
    enddo
!omp$ end parallel do
  endif

  if (myid == np) then
!omp$ parallel do
    do i=0,n
      u(i,je) = (i * h)**2 + 1.
    enddo
!omp$ end parallel do
  endif

!omp$ parallel do
  do j=js,je
    u(0,j) = ((joff + j) * h)**2
    u(n,j) = ((joff + j) * h)**2 + 1.
  enddo
!omp$ end parallel do

end subroutine set_bc

!----------------------------------------------------------------------

subroutine compute_diff(u,unew,n,js,je,diffnorm)

  implicit none
  integer n, js, je
  real    u(0:n,js-1:je+1), unew(0:n,js-1:je+1)
  integer i, j
  real    diffnorm
  real    dnorm[*]
  integer np, myid

  np   = num_images()
  myid = this_image()

  dnorm = 0.0

!omp$ parallel do reduction(+:dnorm)
  do j=js,je
    do i=1,n-1
      dnorm = dnorm + (unew(i,j) - u(i,j))**2
    end do
  end do
!omp$ end parallel do

  sync all

  if (myid == 1) then
    do i=2,np
      dnorm = dnorm + dnorm[i]
    end do
    dnorm = sqrt(dnorm)
  endif

  sync all

  diffnorm = dnorm[1]

end subroutine compute_diff
